<?php

declare(strict_types=1);

namespace Dedi\SyliusCustomerPricingPlugin\Repository\Trait;

use Dedi\SyliusCustomerPricingPlugin\Entity\ChannelPricing\Customer\CustomerChannelPricing;
use Dedi\SyliusCustomerPricingPlugin\Entity\ChannelPricing\Customer\CustomerChannelPricingInterface;
use Dedi\SyliusCustomerPricingPlugin\Entity\ChannelPricing\CustomerGroup\CustomerGroupChannelPricing;
use Dedi\SyliusCustomerPricingPlugin\Entity\ChannelPricing\CustomerGroup\CustomerGroupChannelPricingInterface;
use Dedi\SyliusCustomerPricingPlugin\Entity\ChannelPricing\Organization\OrganizationChannelPricing;
use Dedi\SyliusCustomerPricingPlugin\Entity\ChannelPricing\Organization\OrganizationChannelPricingInterface;
use Dedi\SyliusCustomerPricingPlugin\Entity\Customer\OrganizationInterface;
use Dedi\SyliusCustomerPricingPlugin\Enum\ChannelPricingType;
use Dedi\SyliusCustomerPricingPlugin\Loader\ValueObject\ChannelPricing;
use Doctrine\Common\Collections\ArrayCollection;
use Doctrine\Common\Collections\Collection;
use Doctrine\ORM\NativeQuery;
use Doctrine\ORM\Query\ResultSetMapping;
use Doctrine\ORM\QueryBuilder;
use Sylius\Component\Core\Model\Channel;
use Sylius\Component\Core\Model\CustomerInterface;
use Sylius\Component\Core\Model\ProductVariantInterface;
use Sylius\Component\Customer\Model\CustomerGroupInterface;

trait ChannelPricingRepositoryTrait
{
    abstract public function createListQueryBuilder(string $alias = 'o'): QueryBuilder;

    public function createCustomerPricingListQueryBuilder(string $alias = 'o')
    {
        $queryBuilder = $this->createListQueryBuilder($alias);
        $queryBuilder
            ->leftJoin(Channel::class, 'channel', 'WITH', 'o.channelCode = channel.code')
            ->leftJoin(CustomerChannelPricing::class, 'ccp', 'WITH', 'o.id = ccp.id')
            ->leftJoin('ccp.customer', 'customer', 'WITH', 'ccp.customer = customer.id')
            ->leftJoin(OrganizationChannelPricing::class, 'ocp', 'WITH', 'o.id = ocp.id')
            ->leftJoin('ocp.organization', 'organization', 'WITH', 'ocp.organization = organization.id')
            ->leftJoin(CustomerGroupChannelPricing::class, 'cgcp', 'WITH', 'o.id = cgcp.id')
            ->leftJoin('cgcp.customerGroup', 'customerGroup', 'WITH', 'cgcp.customerGroup = customerGroup.id');

        return $queryBuilder;
    }

    public function findAllForProductVariantByType(ProductVariantInterface $productVariant, string $type): Collection
    {
        $query = $this->createBaseNativeQuery(
            <<<SQL
            SELECT scp.id, scp.channel_code
            FROM sylius_channel_pricing scp
                INNER JOIN sylius_product_variant spv ON spv.id = scp.product_variant_id
            WHERE spv.id = :productVariant AND scp.`type` = :type
            SQL
            ,
            [
                'productVariant' => $productVariant,
                'type' => $type,
            ],
        );

        return $this->getChannelPricingCollectionFromQuery($query);
    }

    public function findAllDefaultsForProductVariant(ProductVariantInterface $productVariant): Collection
    {
        return $this->findAllForProductVariantByType($productVariant, ChannelPricingType::DEFAULT->value);
    }

    public function findAllDefaultsForProductVariantAndChannel(string $channelCode, ProductVariantInterface $productVariant): array
    {
        return $this->createQueryBuilder('o')
            ->where('o.productVariant = :productVariant')
            ->andWhere('o.channelCode = :channelCode')
            ->andWhere('o NOT INSTANCE OF ' . $this->_em->getClassMetadata(CustomerChannelPricingInterface::class)->getName() . ' AND o NOT INSTANCE OF ' . $this->_em->getClassMetadata(OrganizationChannelPricingInterface::class)->getName() . ' AND o NOT INSTANCE OF ' . $this->_em->getClassMetadata(CustomerGroupChannelPricingInterface::class)->getName())
            ->setParameter('productVariant', $productVariant)
            ->setParameter('channelCode', $channelCode)
            ->getQuery()
            ->getResult();
    }

    public function findAllForCustomerForProductVariant(ProductVariantInterface $productVariant, CustomerInterface $customer): Collection
    {
        $query = $this->createBaseNativeQuery(
            <<<SQL
            SELECT scp.id, scp.channel_code
            FROM sylius_channel_pricing scp
                INNER JOIN dedi_customer_pricing_customer_channel_pricing dcpccp ON scp.id = dcpccp.id
            WHERE scp.product_variant_id = :productVariant
              AND dcpccp.customer_id = :customer
              AND scp.`type` = :type
            SQL
            ,
            [
                'productVariant' => $productVariant,
                'customer' => $customer,
                'type' => ChannelPricingType::CUSTOMER->value,
            ],
        );

        return $this->getChannelPricingCollectionFromQuery($query);
    }

    public function findAllForCustomerForProductVariantAndChannel(string $channelCode, ProductVariantInterface $productVariant, CustomerInterface $customer): array
    {
        return $this->createQueryBuilder('o')
            ->innerJoin(CustomerChannelPricing::class, 'ccp', 'WITH', 'o.id = ccp.id')
            ->innerJoin('ccp.customer', 'customer', 'WITH', 'ccp.customer = customer.id')
            ->where('o.productVariant = :productVariant')
            ->andWhere('customer = :customer')
            ->andWhere('o.channelCode = :channelCode')
            ->setParameter('productVariant', $productVariant)
            ->setParameter('customer', $customer)
            ->setParameter('channelCode', $channelCode)
            ->getQuery()
            ->getResult();
    }

    public function findAllForOrganizationForProductVariant(ProductVariantInterface $productVariant, OrganizationInterface $organization): Collection
    {
        $query = $this->createBaseNativeQuery(
            <<<SQL
            SELECT scp.id, scp.channel_code
            FROM sylius_channel_pricing scp
                INNER JOIN dedi_customer_pricing_organization_channel_pricing dcpocp ON scp.id = dcpocp.id
            WHERE scp.product_variant_id = :productVariant
              AND dcpocp.organization_id = :organization
              AND scp.`type` = :type
            SQL
            ,
            [
                'productVariant' => $productVariant,
                'organization' => $organization,
                'type' => ChannelPricingType::ORGANIZATION->value,
            ],
        );

        return $this->getChannelPricingCollectionFromQuery($query);
    }

    public function findAllForOrganizationForProductVariantAndChannel(string $channelCode, ProductVariantInterface $productVariant, OrganizationInterface $organization): array
    {
        return $this->createQueryBuilder('o')
            ->innerJoin(OrganizationChannelPricing::class, 'ocp', 'WITH', 'o.id = ocp.id')
            ->innerJoin('ocp.organization', 'organization', 'WITH', 'ocp.organization = organization.id')
            ->where('o.productVariant = :productVariant')
            ->andWhere('organization = :organization')
            ->andWhere('o.channelCode = :channelCode')
            ->setParameter('productVariant', $productVariant)
            ->setParameter('organization', $organization)
            ->setParameter('channelCode', $channelCode)
            ->getQuery()
            ->getResult();
    }

    public function findAllForCustomerGroupForProductVariant(ProductVariantInterface $productVariant, CustomerGroupInterface $customerGroup): Collection
    {
        $query = $this->createBaseNativeQuery(
            <<<SQL
            SELECT scp.id, scp.channel_code
            FROM sylius_channel_pricing scp
                INNER JOIN sylius_product_variant spv ON spv.id = scp.product_variant_id
                INNER JOIN dedi_customer_pricing_customer_group_channel_pricing dcpcgcp ON scp.id = dcpcgcp.id
            WHERE spv.id = :productVariant
              AND dcpcgcp.customer_group_id = :customerGroup
              AND scp.`type` = :type
            SQL
            ,
            [
                'productVariant' => $productVariant,
                'customerGroup' => $customerGroup,
                'type' => ChannelPricingType::CUSTOMER_GROUP->value,
            ],
        );

        return $this->getChannelPricingCollectionFromQuery($query);
    }

    public function findAllForCustomerGroupForProductVariantAndChannel(string $channelCode, ProductVariantInterface $productVariant, CustomerGroupInterface $customerGroup): array
    {
        return $this->createQueryBuilder('o')
            ->innerJoin(CustomerGroupChannelPricing::class, 'cgcp', 'WITH', 'o.id = cgcp.id')
            ->innerJoin('cgcp.customerGroup', 'customerGroup', 'WITH', 'cgcp.customerGroup = customerGroup.id')
            ->where('o.productVariant = :productVariant')
            ->andWhere('customerGroup = :customerGroup')
            ->andWhere('o.channelCode = :channelCode')
            ->setParameter('productVariant', $productVariant)
            ->setParameter('customerGroup', $customerGroup)
            ->setParameter('channelCode', $channelCode)
            ->getQuery()
            ->getResult();
    }

    private function createBaseNativeQuery(string $sql, array $parameters): NativeQuery
    {
        $em = $this->getEntityManager();

        $rsm = new ResultSetMapping();
        $rsm->addScalarResult('id', 'id', \Doctrine\DBAL\Types\Types::INTEGER);
        $rsm->addScalarResult('channel_code', 'channelCode', \Doctrine\DBAL\Types\Types::STRING);
        $rsm->newObjectMappings['id'] = [
            'className' => ChannelPricing::class,
            'objIndex' => 0,
            'argIndex' => 0,
        ];
        $rsm->newObjectMappings['channel_code'] = [
            'className' => ChannelPricing::class,
            'objIndex' => 0,
            'argIndex' => 1,
        ];
        $query = $em->createNativeQuery($sql, $rsm);

        foreach ($parameters as $key => $value) {
            $query->setParameter($key, $value);
        }

        return $query;
    }

    /**
     * @return Collection<string, ChannelPricing>
     */
    private function getChannelPricingCollectionFromQuery(NativeQuery $query): Collection
    {
        /** @var ChannelPricing[] $result */
        $result = $query->getResult();

        return array_reduce($result, static function (Collection $channelPricings, ChannelPricing $channelPricing) {
            $channelPricings->set($channelPricing->getChannelCode(), $channelPricing);

            return $channelPricings;
        }, new ArrayCollection());
    }
}
